      module input

      public

      integer, parameter         :: bsize = 1000000
      character(bsize), target   :: buffer
      integer                    :: lines(20000)

      integer, parameter  :: xlength = 3000

      integer, parameter  :: IOSTAT_EOF = -1
      integer, parameter  :: IOSTAT_EOR = -2
      integer, save       :: RECL = 1000000

#if defined (FFC) || defined (IFC)
      character(3), save  :: PAD = 'yes'
#else
      character(2), save  :: PAD = 'no'
#endif

      contains

!*******************************************************************************

!*** if buffer is a char array, some nice intrinsics could be used!

!*** when a c stream is used: replace all c(13) and c(13)c(10) with c(10)
!*** one could (but is that to be preferred) to also let that be handled
!*** by the final application; besides - only in CDATA is the space
!*** meaningful; maybe replace it in the cdata parser

      subroutine count_lines(buffer,length,lines)

      implicit none
      character(*), intent(in)  :: buffer
      integer, intent(in)       :: length
      integer, intent(inout)    :: lines(*)
      integer                   :: i,j


      j = 0
      i = 1
      do while (i <= length)
          select case(buffer(i:i))
          case(achar(10))
              j = j + 1
              lines(j) = i
          case(achar(13))
              if (i /= length) then
                  if (buffer(i+1:i+1) == achar(10)) i = i + 1
              endif
              j = j + 1
              lines(j) = i
          case default
              continue
          end select
          i = i + 1
      enddo

      end subroutine count_lines

!*******************************************************************************

      function xline(lines,c) result(x)

      implicit none
      integer, intent(in) :: lines(:),c
      integer             :: x(2)
      integer             :: i


      do i=1,size(lines)
          if (lines(i) >= c) exit
      enddo
      x(1) = i
      x(2) = c - lines(i-1)

      end function xline

!*******************************************************************************

      subroutine open_tstream(file,io,status)


      implicit none
      character(*), intent(in)        :: file
      integer, intent(out)            :: io
      integer, intent(out), optional  :: status
      integer                         :: iostat


      io = 10

!     if (.not. stest) call init_stuff(status)
!     if (status /= 0) stop 'INIT_STUFF error'

      open(io,file=file,action='read',pad=PAD,recl=RECL,iostat=iostat)
      if (present(status)) status = iostat

      end subroutine open_tstream

!*******************************************************************************

      subroutine fill_buffer(io,buffer,status)

      implicit none
      integer, intent(in)        :: io
      character(*), intent(out)  :: buffer
      integer, intent(out)       :: status
      integer                    :: current,i,length,s


      length = len(buffer)
      status = 0
      current = 1

      do while (current <= length)
          s = min(xlength,length - current)
          read(io,'(a)',advance='no',iostat=i,size=s) buffer(current:current+s)
          select case(i)
          case(0)
!         if (i == 0) then
              current = current + s
          case(IOSTAT_EOF)
!         elseif (i == IOSTAT_EOF) then
              current = current - 1 ! remove the added achar(10) just before EOF
              status = current
              return
          case(IOSTAT_EOR)
!         elseif (i == IOSTAT_EOR) then
!*** This assumes only 1 EOR character has been read!!! (s = linelength + 1)
!*** This is correct for 'unix' files, but what about Windows files?
              current = current + s
!write(6,*) '#'//buffer(current-2:current-1)//'#',s,iachar(buffer(current:current)),iachar(buffer(current+1:current+1))
!*** Then 2 characters are read:
!             current = current + s - 1

!*** What when EOR has 2 characters and s falls just inbetween those 2???

              buffer(current:current) = achar(10)
              current = current + 1
          case default
!         else
              status = -1
              return
          end select
!         endif
      enddo
      status = -2

      end subroutine fill_buffer

!*******************************************************************************

      subroutine xtest(status)

      implicit none
      integer, intent(out) :: status
      integer, parameter       :: io = 10
      integer                  :: size,iostat
      character(*), parameter  :: f = '(a)'
      character(40)            :: buffer,inpad
      integer                  :: eor,eof

!-------------------------------------------------------------------------------

      open(io,status='scratch',action='readwrite',pad=inpad)
      write(io,f) 'XXXXXXXX' //              achar(10) //  & ! Unix         LF
                  'YYYYYYYY' // achar(13) // achar(10) //  & ! Windows   CR LF
                  'ZZZZZZZZ' // achar(13) //               & ! Mac       CR
                  'AAAAAAAA'
      endfile(io)
      rewind(io)


      size = 0
      read(io,f,advance='no',size=size,iostat=iostat) buffer(1:)
      if (iostat >= 0) then
          status = -1
          return
      endif
      eor = iostat
      select case(size)
      case(8)
!         LF is an EOR
      case(17)
!         CR is an EOR, LF is read as an ordinary character
!         or
!         CR LF is an EOR, LF is read as an ordinary character
!         or .... ?
      case(0)
          status = -1
          return
      case default
          status = -1
          return
      end select


      size = 0
      read(io,f,advance='no',size=size,iostat=iostat) buffer(1:)
      if (iostat >= 0) then
          status = -1
          return
      endif
      select case(size)
      case(8)
!         LF is an EOR
      case(17)
!         CR is an EOR, LF is read as an ordinary character
!         or
!         CR LF is an EOR, LF is read as an ordinary character
!         or .... ?
      case(0)
          status = -1
          return
      case default
          status = -1
          return
      end select


      size = 0
      read(io,f,advance='no',size=size,iostat=iostat) buffer(1:)
      if (iostat >= 0) then
          status = -1
          return
      endif
      select case(size)
      case(8)
!         LF is an EOR
      case(17)
!         CR is an EOR, LF is read as an ordinary character
!         or
!         CR LF is an EOR, LF is read as an ordinary character
!         or .... ?
      case(0)
          status = -1
          return
      case default
          status = -1
          return
      end select


      size = 0
      read(io,f,advance='no',size=size,iostat=iostat) buffer(1:)
      if (iostat >= 0) then
          status = -1
          return
      endif
      eof = iostat
      if (eor >= 0) then
          status = -1
          return
      endif
      if (eof >= 0) then
          status = -1
          return
      endif
      if (eor == eof) then
          status = -1
          return
      endif

      close(io)
      status = 0
      return


!     Error handling
 1    continue
      close(io)
      status = -1
      return

      end subroutine xtest

!*******************************************************************************

      end module input
